/**
 * Diese Klasse ist die Superklasse fr alle Situationsbeurteilungen.
 * 
 * @author Sebastian Hantsch
 * @version 1.0
 */
public abstract class BasicRating
{
	/** Referenzen zum Weltmodell, Sender und Sprecher */
	protected WorldModel worldModel;
	protected Sender sender;
	protected Speaker speaker;

	/** Startposition des Spielers */
	protected Arrow startPos;
	/** bevorzugte Position whrend des Spiels */
	protected Arrow preferredPos;
	/** Position des gegnerischen Tors */
	protected Arrow enemyGoal = new Arrow(53, 0);
	/** Spieler bewegt zu Beginn */
	protected boolean moved;
	/** Zeitpunkt des letzten turn Befehls im Zusammenhang mit
	 * der Rckkehr ins Spielfeld
	 */
	protected int lastTurnIntoField;

	/**
	 * Erzeugt ein neues <code>BasicRating</code>.
	 */
	public BasicRating(WorldModel worldModel, Sender sender)
	{
		this.worldModel = worldModel;
		this.sender = sender;

		// Sprecher erzeugen
		speaker = new Speaker(worldModel, sender);

		preferredPos = null;
		moved = false;
		lastTurnIntoField = 0;
	}

	/**
	 * Gibt an, ob eine Position fr den Spieler relevant ist.
	 * @return <code>true</code>, falls relevant, sonst <code>false</code>.
	 */
	protected abstract boolean isRelevantPos(Arrow position);

	/**
	 * Beurteilt die aktuelle Spielsituation
	 * @param curAction aktuelle Aktion
	 * @return neue Aktion
	 */
	public abstract BasicAction rate(BasicAction curAction);

	/**
	 * Gibt die Anzahl relevanter Gegner zwischen zwei
	 * Punkten an.
	 * @param fromPos Ausgangsposition
	 * @param toPos Zielposition
	 * @return Anzahl relevanter Gegner
	 */
	private int enemyBetween(Arrow fromPos, Arrow toPos)
	{
		int enemyCount = 0;
		Arrow ePos;
		double toAngle = RoboCupMath.angle(toPos, fromPos);

		// gegnerische Spieler durchgehen
		for (int i = 1; i <= worldModel.getEnemyCount(); i++)
		{
			if (worldModel.getEnemy(i) != null)
			{
				// Ist Gegner in einer relevanten Position
				ePos = worldModel.getEnemy(i).getPosition();
				if (RoboCupMath.distance(ePos, toPos) < 4)
				{
					enemyCount++;
				}
				else if (
					RoboCupMath.distance(fromPos, ePos)
						< RoboCupMath.distance(fromPos, toPos))
				{
					double eAngle = RoboCupMath.angle(ePos, fromPos);
					if (Math.abs(RoboCupMath.angleDiff(toAngle, eAngle))
						< 30.0)
					{
						enemyCount++;
					}
				}
			}
		}
		return enemyCount;
	}

	/**
	 * Ermittelt ein Schussziel fr den Spieler
	 * @param myPos Position des Spielers
	 * @param defPos Ersatzziel, falls kein passendes Ziel
	 * gefunden werden konnte.
	 * @return Schussziel
	 */
	protected Arrow getKickTarget(Arrow myPos, Arrow defPos)
	{
		int plIndex = 0;
		int minEnemy = 100;

		if (worldModel.getPartyCount() < 1)
		{
			// Kein Mitspieler sichtbar, also Ersatzziel nehmen
			return defPos;
		}
		else
		{
			for (int i = 1; i <= worldModel.getPartyCount(); i++)
			{
				Player pp = worldModel.getParty(i);
				if (pp != null)
				{
					// Ist Mitspieler vor einem?
					if (pp.getPosition().getX() > myPos.getX())
					{
						int eCount = enemyBetween(myPos, pp.getPosition());
						if (eCount < minEnemy)
						{
							// Minimale Anzahl an Gegnern zwischen
							// eigener und Zielposition, also merken
							plIndex = i;
						}
					}
				}
			}

			if (plIndex > 0)
			{
				// passender Mitspieler nehmen
				return worldModel.getParty(plIndex).getPosition();
			}
			else
			{
				// Ersatzziel nehmen
				return defPos;
			}
		}
	}

	/**
	 * Gibt an, ob ein eigener Abstoss vorliegt.
	 * @return <code>true</code>, falls eigener Abstoss, sonst
	 * <code>false</code>.
	 */
	public boolean isOwnGoalKick()
	{
		String supposedMsg = "goal_kick_" + worldModel.getPlayDirection();
		return worldModel.getMatchStatus().equals(supposedMsg);
	}

	/**
	 * Gibt an, ob ein eigener Freistoss vorliegt.
	 * @return <code>true</code>, falls eigener Freistoss, sonst
	 * <code>false</code>.
	 */
	public boolean isOwnFreeKick()
	{
		String supposedMsg = "free_kick_" + worldModel.getPlayDirection();
		return worldModel.getMatchStatus().equals(supposedMsg);
	}

	/**
	 * Gibt eine <code>RunAction</code> an.
	 * @return eine <code>RunAction</code>.
	 */
	protected RunAction getRunAction(BasicAction curAction)
	{
		if (curAction != null)
		{
			if (curAction instanceof RunAction)
			{
				return (RunAction) curAction;
			}
			else
			{
				return new RunAction(worldModel, sender);
			}
		}
		else
		{
			return new RunAction(worldModel, sender);
		}
	}

	/**
	 * Gibt eine <code>TurnAction</code> an.
	 * @return eine <code>TurnAction</code>.
	 */
	protected TurnAction getTurnAction(BasicAction curAction)
	{
		if (curAction != null)
		{
			if (curAction instanceof TurnAction)
			{
				return (TurnAction) curAction;
			}
			else
			{
				return new TurnAction(worldModel, sender);
			}
		}
		else
		{
			return new TurnAction(worldModel, sender);
		}
	}

	/**
	 * Gibt eine <code>BallFindAction</code> an.
	 * @return eine <code>BallFindAction</code>.
	 */
	protected BallFindAction getBallFindAction(BasicAction curAction)
	{
		if (curAction != null)
		{
			if (curAction instanceof BallFindAction)
			{
				return (BallFindAction) curAction;
			}
			else
			{
				return new BallFindAction(worldModel, sender);
			}
		}
		else
		{
			return new BallFindAction(worldModel, sender);
		}
	}

	/**
	 * Gibt eine <code>GoalieSlideAction</code> an.
	 * @return eine <code>GoalieSlideAction</code>.
	 */
	protected GoalieSlideAction getGoalieSlideAction(BasicAction curAction)
	{
		if (curAction != null)
		{
			if (curAction instanceof GoalieSlideAction)
			{
				return (GoalieSlideAction) curAction;
			}
			else
			{
				return new GoalieSlideAction(worldModel, sender);
			}
		}
		else
		{
			return new GoalieSlideAction(worldModel, sender);
		}
	}

	/**
	 * Berechnet einen Ballabfangkurs.
	 * @return Abfangposition
	 */
	protected Arrow getBallCollCourse()
	{
		Arrow myPos = worldModel.getYourself().getPosition();
		Arrow bPos = worldModel.getBall().getPosition();
		Arrow bVec = worldModel.getBall().getVector();
		Arrow futurePos = new Arrow(bPos.getX(), bPos.getY());

		double minFactor = 50;
		double distPerCycle = 0.4;

		for (int i = 1; i <= 50; i++)
		{
			futurePos.addArrow(bVec);

			double curDist =
				RoboCupMath.distance(myPos, futurePos)
					- (double) i * distPerCycle;
			if (curDist < 0)
			{
				minFactor = i;
				break;
			}
		}
		return new Arrow(
			bPos.getX() + bVec.getX() * minFactor,
			bPos.getY() + bVec.getY() * minFactor);
	}

	/**
	 * Teilt den anderen Spielern mittels des Sprechers die
	 * Ballposition mit
	 */
	protected void reportBallPos()
	{
		if (worldModel.isballSeen())
		{
			if (worldModel.getActTime() - speaker.getLastMsgTime() > 5)
			{
				speaker.reportBallPos();
			}
		}
	}

	/**
	 * Gibt an, ob eine Spielunterbrechung vorliegt.
	 * @return <code>true</code>, falls Spielunterbrechung, sonst
	 * <code>false</code>.
	 */
	protected boolean isGameBreak()
	{
		String matchStatus = worldModel.getMatchStatus();

		return matchStatus.equals("before_kick_off")
			|| matchStatus.startsWith("goal_l")
			|| matchStatus.startsWith("goal_r");
	}

	/**
	 * Gibt an, ob ein gegnerischer Freistoss vorliegt.
	 * @return <code>true</code>, falls gegnerischer Freistoss, sonst
	 * <code>false</code>.
	 */
	protected boolean isEnemyFreeKick()
	{
		String matchStatus = worldModel.getMatchStatus();
		char enemySide = (worldModel.getPlayDirection() == 'l') ? 'r' : 'l';

		return matchStatus.equals("kick_off_" + enemySide)
			|| matchStatus.equals("kick_in_" + enemySide)
			|| matchStatus.equals("goal_kick_" + enemySide)
			|| matchStatus.equals("corner_kick_" + enemySide)
			|| matchStatus.equals("free_kick_" + enemySide);
	}

	/**
	 * Gibt an, ob der Spieler im Abseits ist.
	 */
	protected boolean isOffside()
	{
		double myX = worldModel.getYourself().getPosition().getX();
		if (worldModel.isballSeen())
		{
			if (worldModel.getBall().getPosition().getX() > myX)
			{
				return false;
			}
		}
		if (worldModel.getEnemyCount() == 0
			&& worldModel.getUnknownsCount() == 0)
		{
			return true;
		}
		else
		{
			if (worldModel.getEnemyCount() > 0)
			{
				int eBehind = 0;
				for (int i = 1; i <= worldModel.getEnemyCount(); i++)
				{
					if (worldModel.getEnemy(i).getPosition().getX() > myX)
					{
						eBehind++;
					}
				}
				return (eBehind < 2);
			}

			if (worldModel.getUnknownsCount() > 0)
			{
				int uBehind = 0;
				for (int i = 1; i <= worldModel.getUnknownsCount(); i++)
				{
					if (worldModel.getUnknowns(i).getPosition().getX() > myX)
					{
						uBehind++;
					}
				}
				return (uBehind < 2);
			}
			return false;
		}
	}

	/**
	 * Lsst den Spieler in das Spielfeld zurckkehren.
	 */
	public void returnIntoField()
	{
		if (worldModel.isLineVisible())
		{
			sender.dash(100);
		}
		else
		{
			double headAngle = worldModel.getYourself().getHeadAngle();
			if (Math.abs(headAngle) > 2)
			{
				sender.turnNeck(-headAngle);
			}
			else
			{
				if (lastTurnIntoField < worldModel.getActTime() - 2)
				{
					sender.turn(90);
					lastTurnIntoField = worldModel.getActTime();
				}
			}
		}
	}
}
