import java.io.*;
import java.net.*;

/**
 * Diese Klasse stellt die grundlegenden Sende- und Empfangsroutinen fr die
 * Kommunikation mit dem RoboCup Server zur Verfgung, basierend auf UDP/IP.
 *
 * 27.5.2002: Empfangspuffergre orientiert sich jetzt an SO_RCVBUF.<br>
 * 03.6.2002: Besseres Abschneiden der Kontrollzeichen beim Empfangen.<br>
 * 25.6.2002: Linuxtauglichkeit verbessert (Beschrnkung der Buffergre).<br>
 *
 * @author Sebastian Hantsch
 * 
 * Basiert auf der gleichnamigen Klasse des SimpleClients von
 * <a href="mailto:kuehne@pi4.informatik.uni-mannheim.de">Gerald Kuehne</a> und
 * <a href="mailto:beier@pi4.informatik.uni-mannheim.de">Markus Beier</a>.
 * 
 * @version 1.2
 */
public class CommunicationChannel
{
	/** <code>DatagramSocket</code> zum Senden und Empfangen von UDP-Packeten */
	private DatagramSocket socket;
	/** IP Adresse des Servers */
	private InetAddress serverAddress;
	/** Port des Servers */
	private int serverPort;
	/** Empfangspuffergre */
	private int bufferSize;

	/**
	 * Erzeugt einen neuen <code>CommunicationChannel</code>.
	 *
	 * @param serverName Die Serveradresse
	 * @param serverPort Der Serverport
	 */
	public CommunicationChannel(String serverName, int serverPort)
	{
		try
		{
			// Neuen DatagramSocket erzeugen
			socket = new DatagramSocket();
			// Serveradresse setzen
			serverAddress = InetAddress.getByName(serverName);
			// Serverport setzen
			this.serverPort = serverPort;
			// Empfangspuffergre setzen
			bufferSize = Math.min(socket.getReceiveBufferSize(), 4096);
		}
		catch (SocketException se)
		{
			System.err.println(
				"CommunicationChannel: Fehler beim Erzeugen des DatagramSockets.");
			se.printStackTrace();
			System.exit(1);
		}
		catch (UnknownHostException uhe)
		{
			System.err.println(
				"CommunicationChannel: Fehler beim Auflsen des Servernamens.");
			uhe.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * Sendet einen <code>String</code> zum Server
	 *
	 * @param message Ein <code>String</code>, der die an den Server zu sendende Nachricht enthlt.
	 */
	public void send(String message)
	{
		try
		{
			// Nachricht in byte-Feld umwandeln
			byte[] buffer = message.getBytes();
			// Neues UDP Packet erzeugen...
			DatagramPacket packet =
				new DatagramPacket(
					buffer,
					buffer.length,
					serverAddress,
					serverPort);
			// ...und an den Server senden
			socket.send(packet);
		}
		catch (IOException ioe)
		{
			System.err.println(
				"CommunicationChannel: Fehler beim Versenden des UDP-Packets.");
			ioe.printStackTrace();
		}
	}

	/**
	 * Empfngt eine Nachricht vom Server (synchron).
	 *
	 * @return Ein <code>String</code> mit der empfangenen Servernachricht.
	 */
	public String receive()
	{
		// Empfangsbuffer anlegen
		byte[] buffer = new byte[bufferSize];
		// Empfangspacket erzeugen...
		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		try
		{
			// ...und eine Nachricht empfangen (blockiert die Ausfhrung!).
			socket.receive(packet);
		}
		catch (IOException ioe)
		{
			System.err.println(
				"CommmunicationChannel: Fehler beim Empfangen eines UDP-Packets.");
			ioe.printStackTrace();
		}
		// Serverport fr die nachfolgenden Sendenvorgnge ermitteln
		serverPort = packet.getPort();
		// Kontrolzeichen abschneiden...
		int truncPos = packet.getLength();
		while (truncPos > 0 && buffer[truncPos - 1] != ')')
		{
			truncPos--;
		}
		// ...und in einen String umwandeln,		
		return new String(buffer).substring(0, truncPos);
	}
}
