import java.util.*;

/**
 * Diese Klasse wertet die gehrten Nachrichten des Spielers
 * aus, sofern diese nicht vom Schiedsrichter ("referee") stammen.
 * 
 * @author Sebastian Hantsch
 * @version 1.1
 */
public class Hearer
{
	/** Referenz zum Weltmodell */
	private WorldModel worldModel;
	/** Der Hashcode des eigenen Teamnamens */
	private int teamHash;
	/** Hashtabelle mit den Nachrichtenklassen */
	private Hashtable classTable;
	/** letzte empfangene <code>BallPosMsg</code>*/
	private BallPosMsg ballPosMsg;
	/** letzte empfangene <code>BallOwnershipMsg</code>*/	
	private BallOwnershipMsg ballOwnershipMsg;
	
	/**
	 * Erzeugt einen neuen <code>Hearer</code>.
	 * @param WorldModel Referenz zum Weltmodell
	 */
	public Hearer(WorldModel worldModel)
	{		
		this.worldModel = worldModel;
		
		// Hashcode des eigenen Teamnamens berechnen
		teamHash = worldModel.getTeamName().hashCode();
		// Nachrichtenklassenhashtabelle erzeugen
		createClassTable();
	}
	
	/**
	 * Erzeugt die Zuordnungstabelle vom Nachrichtennamen zum
	 * zugehrigen Klassennamen.
	 */
	private void createClassTable()
	{
		classTable = new Hashtable();

		try {
			classTable.put("B", Class.forName("BallPosMsg"));
			classTable.put("BO", Class.forName("BallOwnershipMsg"));
		} catch(ClassNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * Diese Methode erzeugt abhngig von der gehrten Nachricht
	 * ein entsprechendes Nachrichtenobjekt mit Hilfe der
	 * Nachrichtenklassenhashtabelle.
	 * 
	 * @param String text Der Nachrichtentext.
	 */
	private void parseHeardText(String text)
	{
		// Nachricht in Teilstcke zerlegen
		ParenthesesTokenizer hTokens = new ParenthesesTokenizer(text.substring(1, text.length() - 1));
		int hash = 0;
		int playerNumber = 0;
		int msgCount = 0;
		int msgTime = 0;

		// Jede Nachricht muss mindestens 5 Teilstcke haben
		if (hTokens.tokenCount() < 5)
		{
			return;
		}

		try
		{
			hash = Integer.parseInt(hTokens.getToken(0));

			// Nachricht vom eigenem Team?	
			if (hash != teamHash)
			{
				return;
			}
		}
		catch (NumberFormatException nfe)
		{
			return;
		}
		
		// Klassennamen fr die Nachricht ermitteln
		HearMsg msg = null;
		String msg_type = hTokens.getToken(4);
		Class className = (Class) classTable.get(msg_type);

		if (className != null)
		{
			// passende Klasse gefunden, also nun Objektinstanz
			// dazu erzeugen
			try
			{
				msg = (HearMsg) className.newInstance();
			}
			catch (Exception e)
			{
				System.err.println("Fehler bei className.newInstance:");
				e.printStackTrace();
			}
		}
		
		if (msg != null)
		{
			// Wenn die Objekterzeugung geklappt hat, die
			// Nachricht verarbeiten
		    if (msg.processTokens(hTokens))
			{
			
			    // Die neue Nachricht der passenden Referenz
			    // zuweisen (es wird von jedem Nachrichtentyp
			    // immer nur die neueste "aufgehoben")
			    if (msg instanceof BallPosMsg)
				{
				    ballPosMsg = (BallPosMsg) msg;
				}
			    else if (msg instanceof BallOwnershipMsg)
				{
				    ballOwnershipMsg = (BallOwnershipMsg) msg;
				}
			}
		}
	}

	/**
	 * Verarbeitung der angefallenen Nachrichten beim Spieler
	 */
	public void processHeard()
	{
		// Alle Nachrichten durchgehen, dabei die Nachrichten,
		// die verarbeitet wurden, entfernen.
		for (int i = 0; i < worldModel.getYourself().getHeardTextCount(); i++)
		{
			parseHeardText(worldModel.getYourself().getHeardMessage(0));
			worldModel.getYourself().clearHeardText(0);
		}
	}
	
	/**
	 * Gibt die letzte empfangene <code>BallOwnershipMsg</code>
	 * zurck und lscht danach die Referenz auf diese, um zu
	 * verhindern, dass die gleiche Nachricht mehrmals als
	 * neu empfangene Nachricht interpretiert wird.
	 * 
	 * @return BallOwnershipMsg 
	 */
	public BallOwnershipMsg getBallOwnershipMsg()
	{
		BallOwnershipMsg msg = ballOwnershipMsg;
		ballOwnershipMsg = null;
		return msg;
	}

	/**
	 * Gibt die letzte empfangene <code>BallPosMsg</code>
	 * zurck und lscht danach die Referenz auf diese, um zu
	 * verhindern, dass die gleiche Nachricht mehrmals als
	 * neu empfangene Nachricht interpretiert wird.
	 *
	 * @return BallPosMsg
	 */
	public BallPosMsg getBallPosMsg()
	{
		BallPosMsg msg = ballPosMsg;
		ballPosMsg = null;
		return msg;
	}
}
