import java.util.Hashtable;

/**
 * Diese Klasse verarbeitet die Servernachrichten, welche vom 
 * <code>CommunicationChannel</code> empfangen wurden.
 * 
 * 26.5.2002: Visualisierungsanbindung
 * 03.6.2002: Neue Methode, um Ereignisklassen zu erzeugen
 *
 * @author <a href="mailto:shantsch@rumms.uni-mannheim.de">Sebastian Hantsch</a>
 * @version 1.2
 */
public class Receiver extends Thread
{
	// Referenzen zu den bentigten Objekten
	private CommunicationChannel commChannel;
	private WorldModel worldModel;
	private Strategy strategy;
	private Visualization visualization;

	// Hashtabelle mit den Ereignisklassen
	private Hashtable classTable;
	// Beendigungsbedingung
	public boolean terminated;

	/**
	 * Erzeugt einen neuen Empfnger.
	 * 
	 * @param commChannel Der zu verwendende <code>CommunicationChannel</code>
	 * @param worldModel Das zu verwendende <code>WorldModel</code>
	 * @param strategy Die zu verwendende <code>Strategy</code>
	 */
	public Receiver(
		CommunicationChannel commChannel,
		WorldModel worldModel,
		Strategy strategy)
	{
		// Referenzen setzen
		this.commChannel = commChannel;
		this.worldModel = worldModel;
		this.strategy = strategy;

		visualization = null;
		terminated = false;

		// Zuordnungstabelle erzeugen
		createClassTable();
	}

	/**
	 * Erzeugt die Zuordnungstabelle vom Nachrichtennamen zum
	 * zugehrigen Klassennamen.
	 */
	private void createClassTable()
	{
		classTable = new Hashtable();

		try
		{
			classTable.put("init", Class.forName("InitEvent"));
			classTable.put("sense_body", Class.forName("SenseBodyEvent"));
			classTable.put("see", Class.forName("SeeEvent"));
			classTable.put("score", Class.forName("ScoreEvent"));
			classTable.put("hear", Class.forName("HearEvent"));
		}
		catch (ClassNotFoundException e)
		{
			e.printStackTrace();
			System.exit(1);
		}
	}

	/**
	 * Diese Methode berschreibt <code>run()</code> von <code>Thread</code>. 
	 * Sie wird ausgefhrt, sobald <code>start()</code> fr den <code>Receiver</code>
	 * aufgerufen wurde.
	 *
	 */
	public void run()
	{
		String msg;

		while (!terminated)
		{

			// Servernachricht empfangen - blockiert die Threadausfhrung
			msg = commChannel.receive();

			// Nachricht zerlegen
			ParenthesesTokenizer msgTokens = new ParenthesesTokenizer(msg);

			if (msgTokens.tokenCount() > 0)
			{
				// Passendes Objekt zum Ereignis erzeugen
				ServerEvent event = null;
				String msg_type = msgTokens.getToken(0);
				Class className = (Class) classTable.get(msg_type);

				if (className != null)
				{
					try
					{
						event = (ServerEvent) className.newInstance();
					}
					catch (Exception e)
					{
						System.err.println("Fehler bei className.newInstance:");
						e.printStackTrace();
					}
				}
				else
				{
					if (msg_type.equals("error"))
					{
						System.err.println(msgTokens.getMsg());
					}
				}

				if (event != null)
				{
					try
					{
						// Tokens weitergeben
						if (event.processTokens(msgTokens))
						{
							// event an worldModel bermitteln
							worldModel.addEvent(event);

							if (event instanceof SenseBodyEvent)
							{
								// Strategie aufrufen
								strategy.execute();
							}

							if (event instanceof SeeEvent)
							{
								if (visualization != null)
								{
									visualization.updateContent();
								}
							}
						}
						else
						{
							System.err.println(
								"Ungltige Nachricht: " + msgTokens.getMsg());
						}
					}
					catch (Exception e)
					{
						System.err.println(
							"Exception in WorldModel, Strategy oder Visualization:");
						e.printStackTrace();
					}
				}
			}
		}
	}

	/**
	 * Setzt die Visualisierung fr diesen Spieler
	 * @param Visualization
	 */
	public void setVisualization(Visualization visualization)
	{
		this.visualization = visualization;
	}
}
